#!/usr/bin/env python
"""
Scummbler v2

    This file is part of Scummbler.

    Scummbler is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Scummbler is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Scummbler.  If not, see <http://www.gnu.org/licenses/>.

Laurence Dougal Myers
www.jestarjokin.net

December 2008 - July 2009

TODO: proper clamping of bytes & words, signed & unsigned
"""

from pyparsing import ParseFatalException, ParseResults, Suppress, Literal, Group
import scummbler_vars

# SCUMM limitations
#  these are set for each game, so these should be moved
#  to an external file.
#  (are these even correct?)
maxlocals = 0x10
maxglobals = 0x31F
maxbits = 0xFF

# Enums etc
RP_WORD = True
RP_BYTE = False

#definedValues = {} # key = alias, value = original text to substitute
#def map_parse_actions(actionMap):
    #for k, v in actionMap.items():
        #k.setParseAction(v)

def gen_label(blockType, lineNum, branchNum=0):
    """ blockType = string, "i" for if, "w" for while, "d" for do/while, "f" for for.
    lineNum = int, the line number of the start of the block (used as a unique ID)
    branchNum = int, current branch (only relevant for if blocks)"""
    return "$" + blockType + str(lineNum) + "b" + str(branchNum)

def resolve_parameter(arg, op, pbits, is_word):
    """ Returns either string bytes of arg, and the opcode modified with parameter bits"""
    global parambits
    # this is a crap hack for known variables
    if 'vartype' in arg or 'knownvar' in arg:
        arg = resolve_var(arg)
        op = op | pbits
    else:
        if is_word:
            arg = to_word_LE(arg.value)
        else:
            arg = to_byte(arg.value)
        
    return arg, op

def resolve_var(tok_in, indirect=False):
    if 'knownvar' in tok_in:
        vartype = "Var"
        varval = scummbler_vars.varNames5map[tok_in.knownvar]
    else:
        vartype = tok_in.vartype
        varval = tok_in.value
    
    # Handle indirect variables
    # Indirect vars are a bit nasty.
    if type(varval) is ParseResults:
        if not vartype is "Var":
            # I don't think this is correct, as I have a script example of
            # if (!Bit[5 + Var[1]]) {
            # but does that mean Var[1] is this original indirect thing?
            #raise ParseFatalException("Only Var variables can have indirect pointers as the index.")
            pass
        baseval = varval.base
        
        if "vartype" in varval.offset or "knownvar" in varval.offset:
            offval = resolve_var(varval.offset, True)
        else:
            offval = to_word_LE(varval.offset.value)
        
        # This is crap, I've just duplicated things
        if vartype is "Local":
            vv = (int(baseval) & 0xFF)
            vv = vv | 0x4000
        elif vartype is "Bit":
            vv = int(baseval) & 0x7FFF
            vv = vv | 0x8000
        elif vartype is "Var":
            vv = int(baseval) & 0x1FFF
        
        return to_word_LE(vv, 0x2000) + offval
    
    # NOTE: > may need to be >=, not sure if max value limits are inclusive
    # I've removed the max bounds checking for now because I'm not sure if the max vals are correct
    if vartype is "Local":
        #if int(varval) >= maxlocals:
        #    raise ParseFatalException("Local variable index is too high: " + varval + " (max: " + str(maxlocals) + ")")
        vv = (int(varval) & 0xFF)
        vv = vv | 0x4000
        if indirect:
            vv = vv | 0x2000
        return to_word_LE(vv)
    elif vartype is "Bit":
        #if int(varval) >= maxbits:
        #    raise ParseFatalException("Bit variable index is too high: " + varval + " (max: " + str(maxbits) + ")")
        if indirect:
            #raise ParseFatalException("Bit variables cannot be used in an indirect variable pointer.")
            vv = vv | 0x2000
        vv = int(varval) & 0x7FFF
        vv = vv | 0x8000
        return to_word_LE(vv)
    elif vartype is "Var":
        #if int(varval) >= maxglobals:
        #    raise ParseFatalException("Var variable index is too high: " + varval + " (max: " + str(maxglobals) + ")")
        vv = int(varval) & 0x1FFF
        if indirect:
            vv = vv | 0x2000
        return to_word_LE(vv)
    else:
        raise Exception("Unknown vartype: " + str(vartype) + ", val: " + str(varval))

def to_byte(inval, maskval=None, clamp=True):
    """ Accepts an int or a string representing the value ("123"),
    returns a 1-char string with actual value("\x7B").
    
    Optionally accepts a "mask" value which will be binary-ORed with the value.
    Also accepts a "clamp" boolean (defaults to True), which will make sure the
    input value is between -128 and 255 (doesn't know if value should be
    signed)."""
    if type(inval) is str:
        inval = int(inval)
    if clamp and inval < -128 or inval > 255:
        raise ParseFatalException("Byte value should be between -128 and 255.")
    if maskval != None:
        inval = inval | maskval
    return chr(inval & 0xFF)

def to_word_LE(inval, maskval=None, clamp=True):
    """ Accepts an int or a string representing the value ("123"),
    returns a 2-char string with actual value in BE format ("\x7B\x00").
    
    Optionally accepts a "mask" value which will be binary-ORed with the value.
    Also accepts a "clamp" boolean (defaults to True), which will make sure the
    input value is between -32,768 and 65535 (doesn't know if value should be
    signed)."""
    if type(inval) is str:
        inval = int(inval)
    if clamp and inval < -32768 or inval > 65535:
        raise ParseFatalException("Word value should be between -32768 and 65535.")
    if maskval != None:
        inval = inval | maskval
    return chr(inval & 0xFF) + \
          chr((inval & 0xFF00) >> 8)

def SupLit(instr):
    """ Just tidies up the code a bit, alias for Suppress(Literal(instr))"""
    return Suppress(Literal(instr))

def NamedElement(element, name):
    """ Workaround deficiency in PyParsing, where setting the Results Name of an element
    returns a copy, which interferes with Parse Actions set after the use of that element
    in a compound statement.
    
    NOTE: Actually escapes from the group and returns the first item of element (to escape
    from the Parse Results object and return the actual token).
    I do this because only leString uses this wrapper, but it may cause problems with anything else.
    """
    return Group(element)(name).setParseAction(lambda s, loc, toks: toks[0][0]) # escape from group, then escape from parse results

class IfBlockInfo(object):
    def __init__(self, lineNum):
        """ Accepts the line number starting the "if" block, immediately calls self.pushBranch()"""
        self.ID = lineNum
        self.branchStack = []
        self.branchCounter = 0
        self.pushBranch(lineNum) # start of "if" block is implicitly a branch.
    
    def __iter__(self):
        return iter(self.branchStack)
    
    def __str__(self):
        return "[IfBlock ID: " + str(self.ID) + ",  branchCounter: " + str(self.branchCounter) + ", branchStack: " + str(self.branchStack) + "]"
        
    def pushBranch(self, srcLineNum):
        """ Accepts the line number containing the branch, stores the line number with a new, automatically
        generated label for the destination."""
        self.branchStack.append( (srcLineNum, gen_label('i', self.ID, self.branchCounter)) )
        self.branchCounter += 1
    
    def popElseBranch(self):
        """ Returns the second-last added branch."""
        return self.branchStack.pop(-2)
    
    def popBranch(self, i=-1):
        """ Not used, popElseBranch is more specific."""
        return self.branchStack.pop(i)

class LoopBlockInfo(object):
    def __init__(self, lineNum, incremOp=None):
        """ Accepts the line number starting the loop block, and an optional operation to perform at the end of a "for" block."""
        self.ID = lineNum
        self.startLine = lineNum #public
        self.incremOp = incremOp
        
    def __str__(self):
        return "[LoopBlock ID: " + str(self.ID) + ", endLine: " + str(self.endLine) + "]"
    
class ScummblerException(Exception):
    pass
    
class ScummblerAutogenException(ScummblerException):
    pass

# thrown whenever a parse action is being generated for a non-existing grammar.
class ScummblerAutogenSubopcodeException(ScummblerAutogenException):
    pass

class ScummblerParseException(ScummblerException):
    pass
